/*****************************************************************************
 * Description: Plays the files                                              *
 * Copyright:   2007 by Alexander Motzkau                                    *
 * License:     GPL V2                                                       *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ivorbiscodec.h>
#include <ivorbisfile.h>
#include <mad.h>
#include <ao.h>

#include "player.h"
#include "playlist.h"

class PositionEvent : public QCustomEvent
{
    public:
        PositionEvent(int p) : QCustomEvent(Player::EvPosition),
	    pos(p) { }
	
        int position() const { return pos; }

    private:
        int pos;
};

class NameChangeEvent : public QCustomEvent
{
    public:
        NameChangeEvent(const QString& n) : QCustomEvent(Player::EvNameChange),
	    name(n) {}
	
        const QString& newName() const { return name; };

    private:
        QString name;
};

class SongEndedEvent : public QCustomEvent
{
    public:
        SongEndedEvent() : QCustomEvent(Player::EvSongEnded) {}
};


Player::Player() : QObject(), thread(this), currentfile()
{
    thread.start();
}

Player::~Player()
{
}

void Player::stop()
{
    thread.stop();
    emit position(0);
}

void Player::playpause()
{
    if(currentfile.isNull())
	emit songEnded();
    thread.playpause(currentfile.local8Bit());
}

void Player::select(PlaylistEntry* file, bool waitforend)
{
    if(!waitforend && file)
    {
	if(file->getFileName() != currentfile)
	    stop();
	else
	    return;
    }

    if(file)
    {
	currentfile=file->getFileName();
	thread.next(currentfile.local8Bit());
    }
    else
    {
	currentfile=QString();
	stop();
    }
}

void Player::select(PlaylistEntry* file)
{
    select(file, false);
}

void Player::seek(int promille)
{
    thread.seek(promille);
}

bool Player::event(QEvent * e)
{
    switch(e->type())
    {
	case EvPosition:
	{
	    PositionEvent *pe = (PositionEvent*)e;
	    emit position(pe->position());
	    return true;
	}

	case EvNameChange:
	{
	    NameChangeEvent *ne = (NameChangeEvent*)e;
	    emit changeName(ne->newName());
	    return true;
	}

	case EvSongEnded:
	    emit songEnded();
	    return true;
	
	default:
	    return QObject::event(e);
    }
}

void Player::selectNoGain()
{
    thread.selectGain(PlayerThread::NoGain);
}

void Player::selectTrackGain()
{
    thread.selectGain(PlayerThread::TrackGain);
}

void Player::selectAlbumGain()
{
    thread.selectGain(PlayerThread::AlbumGain);
}

PlayerThread::PlayerThread(Player *pl) : QThread(), player(pl),
    mutex(), cond(), shouldstop(false), shouldpause(false),
    notplaying(false), shouldseek(-1)
{
    nextname[0] = 0;
    nextname[sizeof(nextname)-1] = 0;
}

PlayerThread::~PlayerThread() {}

void PlayerThread::run()
{
    ao_sample_format current_format;
    ao_device *device = NULL;
    
    ao_initialize();
    current_format.byte_format = AO_FMT_NATIVE;
    current_format.bits = 16;
    current_format.rate = 0;
    
    mutex.lock();
    while(true)
    {
	// Outer loop for the songs.
	if(!nextname[0])
	{
	    notplaying = true;
	    cond.wait(&mutex);
	}
	else
	{
	    unsigned long total_length = 1;
	    unsigned long current_pos = 0;
	    int status = 0; // 0: Load file, 1: vf is active, 2: nothing

	    char fname[sizeof(nextname)];
	    char* ending;
	    
	    memcpy(fname, nextname, sizeof(nextname));
	    nextname[0]=0;
	    notplaying = false;
	    
	    ending = fname + strlen(fname) - 4;
	    if(!strncasecmp(ending, ".ogg", 4))
	    {
		OggVorbis_File vf;
		int trackgain_scale;
		int albumgain_scale;
		
		while(!shouldstop)
		{
		    if(shouldpause)
			cond.wait(&mutex);
		    else
		    {
			int seektmp = shouldseek;
			int last_stream;
		        int current_stream;
			GainSelection gainsel = whichgain;
		    
			shouldseek = -1;
		    
		        mutex.unlock();
			// Let's play it...
		    
			if(!status)
			{
			    FILE *in;
			
			    if((in=fopen(fname, "rb"))==NULL ||
				ov_open(in, &vf, NULL, 0) < 0)
			    {
				if(in==NULL)
				    fclose(in);
			    
			        status = 2;
				mutex.lock();
				break;
			    }
			
			    total_length = ov_pcm_total(&vf, -1);
			    status = 1;
			    last_stream = -1;
			}

			if(seektmp>=0)
			    ov_pcm_seek(&vf, ((unsigned long long)seektmp)*total_length/1000);
		    
			long ret=ov_read(&vf,pcmdata,sizeof(pcmdata),&current_stream);
		        if(ret<=0)
			{
			    /* EOF or error in the stream. */
			    QApplication::postEvent(player, new SongEndedEvent());
			    mutex.lock();
			    break;
		        }
			else
		        {
			    unsigned long newpos;
			    int gainscale;
			    
			    if(current_stream != last_stream)
			    {
				vorbis_info *vi = ov_info(&vf, current_stream);
				vorbis_comment *vc = ov_comment(&vf, current_stream);
				
				if(vi->rate != current_format.rate ||
				   vi->channels != current_format.channels)
				{
				    if(current_format.rate)
					ao_close(device);
				
			    	    current_format.rate = vi->rate;
				    current_format.channels = vi->channels;
			    	    device = ao_open_live(ao_default_driver_id(),
					&current_format, NULL);
			    	    if(device == NULL)
				    {
					current_format.rate = 0;
				        mutex.lock();
					break;
				    }
				}
				
				if(vc)
				{
				    double album_gain=0, track_gain=0, album_peak=0, track_peak=0;
				    bool track_set=false, album_set=false;
				    char **comment;
				    
				    for(comment=vc->user_comments; *comment; comment++)
					if(!strncasecmp(*comment, "REPLAYGAIN_TRACK_GAIN=", 22))
					{
					    track_gain = pow(10, atof(*comment+22)/20);
					    track_set = true;
					}
					else if(!strncasecmp(*comment, "REPLAYGAIN_ALBUM_GAIN=", 22))
					{
					    album_gain = pow(10, atof(*comment+22)/20);
					    album_set = true;
					}
					else if(!strncasecmp(*comment, "REPLAYGAIN_TRACK_PEAK=", 22))
					    track_peak = atof(*comment+22);
					else if(!strncasecmp(*comment, "REPLAYGAIN_ALBUM_PEAK=", 22))
					    album_peak = atof(*comment+22);
				
				    if(track_set)
				    {
					/* Anti-Clipping, let's clip instead.
					if(track_gain * track_peak > 1.0)
					    track_gain = 1.0 / track_peak;
					*/
					if(track_gain > 15.0)
					    track_gain = 15.0;
					
					trackgain_scale = int(track_gain * 64.0);
				    }
				    else
					trackgain_scale = 64;

				    if(album_set)
				    {
					/*
					if(album_gain * album_peak > 1.0)
					    album_gain = 1.0 / album_peak;
					*/
					if(album_gain > 15.0)
					    album_gain = 15.0;
					
					albumgain_scale = int(album_gain * 64.0);
				    }
				    else
					albumgain_scale = trackgain_scale;
				}
				
				last_stream = current_stream;
			    }
			    
			    switch(gainsel)
			    {
				case AlbumGain:
				    gainscale = albumgain_scale;
				    break;
				case TrackGain:
				    gainscale = trackgain_scale;
				    break;
				default:
				    gainscale = 64;
			    }
			    
			    if(gainscale != 64)
			    {
				signed short int *buf = (signed short int*) pcmdata;
				
				for(int i=ret/2;i--;buf++)
				{
				    int val = (int(*buf)*gainscale)>>6;

				    if(val > SHRT_MAX)
					*buf = SHRT_MAX;
				    else if(val < SHRT_MIN)
					*buf = SHRT_MIN;
				    else
					*buf = (signed short int)val;
				}
			    }
			
			    ao_play(device, pcmdata, ret);
			    newpos = ov_pcm_tell(&vf)*1000ULL/total_length;
			    if(newpos != current_pos)
			    {
				current_pos = newpos;
				QApplication::postEvent(player, new PositionEvent(current_pos));
			    }
		        }
		    
			mutex.lock();
		    }
	        }
	    
		mutex.unlock();
	    
	        if(status==1)
		    ov_clear(&vf);
	    }
	    else if(!strncasecmp(ending, ".mp3", 4))
	    {
		struct mad_stream stream;
		struct mad_frame  frame;
	        struct mad_synth  synth;
		signed short int *pcm_buf = NULL;
		FILE *in = NULL;

		if(current_format.rate)
		    pcm_buf = (short int*)malloc(
			1152*current_format.channels*sizeof(short int));

		while(!shouldstop)
		{
		    if(shouldpause)
			cond.wait(&mutex);
		    else
		    {
			unsigned long newpos;
			int seektmp = shouldseek;
			mad_fixed_t** pcm_ptrs;
			int pcm_pos;
		    
			shouldseek = -1;
		    
		        mutex.unlock();
			// Let's play it...
			
			if(!status)
			{
			    if((in=fopen(fname, "rb"))==NULL)
			    {
			        status = 2;
				mutex.lock();
				break;
			    }
			
			    fseek(in, 0, SEEK_END);
			    total_length = ftell(in);
			    fseek(in, 0, SEEK_SET);

			    status = 1;
			    
			    mad_stream_init(&stream);
			    mad_frame_init(&frame);
			    mad_synth_init(&synth);
		        }
			
			if(seektmp>=0)
			{
			    mad_stream_finish(&stream);
			    mad_stream_init(&stream);
			    fseek(in, ((unsigned long long)seektmp)*total_length/1000, SEEK_SET);
			}
			
			if(!stream.buffer || stream.error == MAD_ERROR_BUFLEN)
			{
			    size_t remaining = stream.bufend-stream.next_frame;
			    memmove(filedata, stream.next_frame, remaining);
			    
			    size_t len = fread(filedata+remaining, 1, sizeof(filedata)-remaining, in);
			    
			    if(len==0)
			    {
				/* EOF or error in the stream. */
			        QApplication::postEvent(player, new SongEndedEvent());
				mutex.lock();
			        break;
		    	    }
			
			    mad_stream_buffer(&stream, filedata, remaining+len);
			    stream.error = MAD_ERROR_NONE;
			}
			
			if(mad_frame_decode(&frame, &stream))
			{
			    if(MAD_RECOVERABLE(stream.error) ||
			       stream.error == MAD_ERROR_BUFLEN)
			    {
				mutex.lock();
				continue;
			    }
			    else
			    {
				/* EOF or error in the stream. */
			        QApplication::postEvent(player, new SongEndedEvent());
				mutex.lock();
			        break;
		    	    }
			}
			
			mad_synth_frame(&synth, &frame);
			
			if((unsigned)current_format.rate != synth.pcm.samplerate ||
			   current_format.channels != synth.pcm.channels)
			{
			    if(current_format.rate)
			    {
				ao_close(device);
				free(pcm_buf);
			    }
			    
			    current_format.rate = synth.pcm.samplerate;
			    current_format.channels = synth.pcm.channels;
			    device = ao_open_live(ao_default_driver_id(), &current_format, NULL);
			    
			    if(device == NULL)
			    {
				current_format.rate = 0;
				free(pcm_buf);
			        mutex.lock();
			        break;
			    }
			    
			    pcm_buf = (short int*) malloc(
				1152*synth.pcm.channels*sizeof(short int));
			}
			
			pcm_ptrs = (mad_fixed_t**) alloca(
					sizeof(mad_fixed_t*)*synth.pcm.channels);
			
			for(int c=0; c<synth.pcm.channels; c++)
			    pcm_ptrs[c] = synth.pcm.samples[c];
			
			pcm_pos = 0;
			for(int i=synth.pcm.length; i--;)
			{
			    mad_fixed_t** ch = pcm_ptrs;
			    for(int c=synth.pcm.channels; c--;)
			    {
				mad_fixed_t sample = *((*(ch++))++) + (1L << (MAD_F_FRACBITS-16));
				
				if(sample >= MAD_F_ONE)
				    sample = MAD_F_ONE - 1;
				else if(sample < -MAD_F_ONE)
				    sample = -MAD_F_ONE;
				
				pcm_buf[pcm_pos++] = (sample >> (MAD_F_FRACBITS - 15));
			    }
			}
			
			ao_play(device, (char*)pcm_buf, pcm_pos*2);

			newpos = ftell(in)*1000ULL/total_length;
			if(newpos != current_pos)
			{
			    current_pos = newpos;
			    QApplication::postEvent(player, new PositionEvent(current_pos));
			}

			mutex.lock();
		    }
	        }
	    
		mutex.unlock();
	    
	        if(status==1)
		{
		    mad_synth_finish(&synth);
		    mad_frame_finish(&frame);
		    mad_stream_finish(&stream);
		    fclose(in);
		}
		
		if(current_format.rate)
		    free(pcm_buf);
	    }
	    else
	    {
		mutex.unlock();
		QApplication::postEvent(player, new SongEndedEvent());
	    }
		
	    mutex.lock();
	    shouldstop = false;
	    shouldpause= false;
	    shouldseek = -1;
	}
    }
}

void PlayerThread::stop()
{
    mutex.lock();
    shouldstop=true;
    shouldpause=false;
    shouldseek=-1;
    nextname[0]=0;
    cond.wakeAll();
    mutex.unlock();
}

void PlayerThread::playpause(const char* fname)
{
    mutex.lock();

    if(shouldpause)
	shouldpause=0;
    else if(notplaying)
	strncpy(nextname, fname, sizeof(nextname)-1);
    else
	shouldpause=1;

    cond.wakeAll();
    mutex.unlock();
}

void PlayerThread::next(const char* fname)
{
    mutex.lock();
    
    strncpy(nextname, fname, sizeof(nextname)-1);
    
    cond.wakeAll();
    mutex.unlock();
}

void PlayerThread::seek(int pos)
{
    mutex.lock();
    
    shouldseek = pos;
    
    cond.wakeAll();
    mutex.unlock();
}

void PlayerThread::selectGain(GainSelection sel)
{
    mutex.lock();
    
    whichgain = sel;
    
    mutex.unlock();
}
