/*****************************************************************************
 * Description: Handles the playlist                                         *
 * Copyright:   2007 by Alexander Motzkau                                    *
 * License:     GPL V2                                                       *
 *****************************************************************************/

#include <qfile.h>
#include <qtextstream.h>

#include "playlist.h"

PlaylistEntry::PlaylistEntry(const QString& fname) : filename(fname)
{
}
    
QString PlaylistEntry::getFileName() const
{
    return filename;
}

void PlaylistEntry::setFileName(QString str)
{
    filename = str;
}

Playlist::Playlist()
    : QObject(), playlist(), current(-1)
{
}

Playlist::~Playlist() {}
    
void Playlist::addEntry(PlaylistEntry* entry)
{
    playlist.push_back(entry);
    emit entryAdded(playlist.size()-1, entry);
}

void Playlist::delEntry(PlaylistEntry* entry)
{
    for(PLIterator it = playlist.begin();it != playlist.end(); it++)
	if(*it == entry)
	{
	    int pos = it-playlist.begin();
	    playlist.erase(it);
	    emit entryRemoved(pos, entry);
	    delete entry;
	    break;
	}
}

void Playlist::delPos(int pos)
{
    PlaylistEntry* entry = playlist[pos];
    playlist.erase(playlist.begin()+pos);
    emit entryRemoved(pos, entry);
    delete entry;

    if(pos == current)
    {
	if((unsigned)pos == playlist.size())
	    selectPos(pos-1);
	else
	    selectPos(pos);
    }
}

void Playlist::clear()
{
    selectPos(-1);
    for(PLIterator it = playlist.begin();it != playlist.end(); it++)
    {
        emit entryRemoved(0, (*it));
        delete (*it);
    }
    playlist.clear();
}

PlaylistEntry* Playlist::currentEntry() const
{
    return playlist.at(current);
}

void Playlist::selectPos(int pos)
{
    current = pos;
    emit selected(current, (current<0) ? NULL : playlist[current]);
    emit selected((current<0) ? NULL : playlist[current]);
}

void Playlist::selectEntry(PlaylistEntry* entry)
{
    current = getPosByEntry(entry);
    emit selected(current, (current<0) ? NULL : entry);
    emit selected((current<0) ? NULL : entry);
}

void Playlist::movePos(int oldpos, int newpos)
{
    PlaylistEntry* entry = playlist[oldpos];
    
    playlist.insert(playlist.begin()+newpos, entry);
    playlist.erase(playlist.begin()+oldpos+((newpos<oldpos)?1:0));
    
    emit entryMoved(oldpos, entry, newpos);
}

int Playlist::currentPos()
{
    return current;
}

int Playlist::getPosByEntry(PlaylistEntry* entry)
{
    for(PLIterator it = playlist.begin();it != playlist.end(); it++)
	if((*it) == entry)
	    return it-playlist.begin();
    return -1;
}

PlaylistEntry* Playlist::getEntryByPos(int pos)
{
    return (pos<0) ? NULL : playlist[pos];
}

void Playlist::savePlaylist(const QString& fname)
{
    QFile fout(fname);
    if(fout.open(IO_WriteOnly))
    {
        QTextStream out(&fout);
	for(PLIterator it = playlist.begin();it != playlist.end(); it++)
	    out << (*it)->getFileName() << "\n";
	fout.close();
    }
}

void Playlist::addPlaylist(const QString& fname)
{
    QFile fin(fname);
    if(fin.open(IO_ReadOnly))
    {
        QTextStream in(&fin);
        while (!in.eof())
	{
	    QString str = in.readLine();
	    if(!str.isEmpty() && !str.startsWith("#"))
		addEntry(new PlaylistEntry(str));
        }
        fin.close();
    }
}
