/*****************************************************************************
 * Description: GUI of the Vorbis Player                                     *
 * Copyright:   2007 by Alexander Motzkau                                    *
 * License:     GPL V2                                                       *
 *****************************************************************************/

#include <qaction.h>
#include <qaccel.h>
#include <qmenubar.h>
#include <qmessagebox.h>
#include <qtoolbar.h>
#include <qvbox.h>
#include <qslider.h>

#include <opie2/oapplication.h>
#include <opie2/oresource.h>
#include <opie2/oapplicationfactory.h>
#include <opie2/ofiledialog.h>
#include <opie2/oconfig.h>
#include <opie2/oglobal.h>
#include <opie2/odevice.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/fcntl.h>
#include <sys/soundcard.h>
#include <alsa/asoundlib.h>

#include "playlist.h"
#include "player.h"
#include "qtPlayer.h"
#include "qtPlaylist.h"
#include "selector.h"

OPIE_EXPORT_APP( Opie::Core::OApplicationFactory<MainWindow> )

MainWindow::MainWindow(QWidget *parent,  const char* name, WFlags fl )
    : QMainWindow( parent, name, fl ), screenSaver(0), timeCount(0),
    keybgrabbed(false)
{
    setCaption(tr("Vorbis Player") );
    setIcon( Opie::Core::OResource::loadPixmap("oggplay/oggplay", Opie::Core::OResource::SmallIcon ));
    
    playlist = new Playlist();
    player = new Player();
    selector = new Selector();

    initUI();

    QVBox *wspace = new QVBox(this);
    QSlider *slider = new QSlider(wspace);
    slider->setOrientation(QSlider::Horizontal);
    slider->setTracking(true);
    slider->setTickmarks(QSlider::Below);
    slider->setRange(0,1000);
    slider->setLineStep(20);
    slider->setPageStep(200);
    slider->setTickInterval(50);

    qtPlaylist *qplaylist = new qtPlaylist(wspace);
    setCentralWidget( wspace );
    
    connect(playlist, SIGNAL(selected(int,PlaylistEntry*)),
            qplaylist, SLOT(select(int,PlaylistEntry*) ));
    connect(playlist, SIGNAL(entryAdded(int,PlaylistEntry*)),
            qplaylist, SLOT(updateEntry(int,PlaylistEntry*) ));
    connect(playlist, SIGNAL(entryChanged(int,PlaylistEntry*)),
            qplaylist, SLOT(updateEntry(int,PlaylistEntry*) ));
    connect(playlist, SIGNAL(entryRemoved(int,PlaylistEntry*)),
            qplaylist, SLOT(removeEntry(int,PlaylistEntry*) ));
    connect(playlist, SIGNAL(entryMoved(int,PlaylistEntry*,int)),
            qplaylist, SLOT(moveEntry(int,PlaylistEntry*,int) ));

    connect(qplaylist, SIGNAL(itemClicked(int)),
	    playlist, SLOT(selectPos(int)));
    connect(qplaylist, SIGNAL(itemRemove(int)),
	    playlist, SLOT(delPos(int)));
    connect(qplaylist, SIGNAL(itemMove(int,int)),
	    playlist, SLOT(movePos(int,int)));

    connect(playlist, SIGNAL(selected(PlaylistEntry*)),
	    player, SLOT(select(PlaylistEntry*)));
	    
    connect(player, SIGNAL(position(int)),
	    slider, SLOT(setValue(int)));
    connect(slider, SIGNAL(sliderMoved(int)),
	    player, SLOT(seek(int)));

    connect(selector, SIGNAL(choose(PlaylistEntry*,bool)),
	    this, SLOT(nextSong(PlaylistEntry*,bool)));
    connect(selector, SIGNAL(choose(int,bool)),
	    this, SLOT(nextSong(int,bool)));

    connect(playlist, SIGNAL(entryAdded(int, PlaylistEntry*)),
	    selector, SLOT(entryAdded(int, PlaylistEntry*)));
    connect(playlist, SIGNAL(entryRemoved(int, PlaylistEntry*)),
	    selector, SLOT(entryRemoved(int, PlaylistEntry*)));
    connect(playlist, SIGNAL(entryMoved(int, PlaylistEntry*, int)),
	    selector, SLOT(entryMoved(int, PlaylistEntry*, int)));
    connect(playlist, SIGNAL(selected(int, PlaylistEntry*)),
	    selector, SLOT(songSelected(int, PlaylistEntry*)));

    connect(player, SIGNAL(songEnded()),
	    this, SLOT(songEnded()));

    startTimer(100);
    
    playlist->addPlaylist(Opie::Core::OGlobal::applicationFileName("oggplay","playlist.m3u"));
}

MainWindow::~MainWindow()
{
    playlist->savePlaylist(Opie::Core::OGlobal::applicationFileName("oggplay","playlist.m3u"));
}

/*
 * set Document is a special function used by Document
 * centric applications.
 * In example if Opie receives a Todo via IrDa it uses
 * setDocument via QCOP the IPC system to load the card
 * Or If you decide to open a file in filemanager with textedit
 * setDocument is called via IPC in textedit.
 * Also any call to QPE/Application/xyz and xyz is currently not running
 * leads to the start of xyz and delivering of the QCOP call
 * But more on QCOP in the next application
 */
void MainWindow::setDocument( const QString& )
{
}

void MainWindow::nextSong(PlaylistEntry* entry, bool waitforend)
{
    player->select(entry, waitforend);
    playlist->selectEntry(entry);
}

void MainWindow::nextSong(int playlistpos, bool waitforend)
{
    nextSong(playlist->getEntryByPos(playlistpos), waitforend);
}

void MainWindow::songEnded() { selector->forward(true); }
void MainWindow::forward() { selector->forward(false); }
void MainWindow::backward() { selector->backward(false); }

void MainWindow::addToPlaylist()
{
    QMap<QString, QStringList> mimeTypes;
    QStringList musicTypes;
    musicTypes << "application/ogg" << "audio/x-ogg" << "audio/mpeg" << "audio/x-mpegurl";
    mimeTypes.insert( tr(" Music files"), musicTypes );
    mimeTypes.insert( tr("All"), QStringList() );
    mimeTypes.insert( tr("Playlists"), "audio/x-mpegurl" );
    musicTypes = QStringList();
    musicTypes << "application/ogg" << "audio/x-ogg";
    mimeTypes.insert( tr("Ogg Vorbis"), musicTypes );
    mimeTypes.insert( tr("MP3"), "audio/mpeg" );

    QString fname = Opie::Ui::OFileDialog::getOpenFileName(
	Opie::Ui::OFileSelector::EXTENDED,
	QString::null, QString::null, mimeTypes, 0, tr("Add to playlist"));

    if(!fname.isNull())
    {
	QString ending = fname.right(4).lower();
	if(ending == ".ogg" || ending == ".mp3")
	    playlist->addEntry(new PlaylistEntry(fname));
	else if(ending == ".m3u")
	    playlist->addPlaylist(fname);
	else
	    QMessageBox::information(this, tr("Unsupported file format"),
		tr("This file format is not supported."));
    }
}

void MainWindow::addDirToPlaylist()
{
    QString fname = Opie::Ui::OFileDialog::getDirectory(
	Opie::Ui::OFileSelector::EXTENDED,
	QString::null, 0, tr("Add to playlist"));

    if(!fname.isNull())
	addDir(QDir(fname));
}

void MainWindow::addDir(const QDir& dir)
{
    const QFileInfoList *list = dir.entryInfoList(
	QDir::Dirs | QDir::Files | QDir::Readable, 
	QDir::Name | QDir::DirsFirst);
    QFileInfoListIterator it(*list);
    QFileInfo *file;
    
    while((file = it.current()))
    {
	if(file->isDir())
	{
	    if(file->fileName() != "." && file->fileName() != "..")
	        addDir(QDir(file->filePath()));
	}
	else if(file->isReadable())
	{
	    QString ending = file->extension(false).lower();
	    if(ending == "ogg" || ending == "mp3")
		playlist->addEntry(new PlaylistEntry(file->filePath()));
	    else if(ending == "m3u")
		playlist->addPlaylist(file->filePath());
	}
	
	++it;
    }
}

void MainWindow::savePlaylist()
{
    QMap<QString, QStringList> mimeTypes;
    mimeTypes.insert( tr("All"), QStringList() );
    mimeTypes.insert( tr("Playlists"), "audio/x-mpegurl" );
    QString fname = Opie::Ui::OFileDialog::getSaveFileName(
	Opie::Ui::OFileSelector::EXTENDED,
	QString::null, QString::null, mimeTypes, 0, tr("Save playlist"));
    
    if(!fname.isNull())
    {
	QString ending = fname.right(4).lower();
	if(ending == ".m3u")
	    playlist->savePlaylist(fname);
	else
	    QMessageBox::information(this, tr("Unsupported file format"),
		tr("This file format is not supported."));
    }
}

void MainWindow::initUI()
{
    int gaincfg;
    
    //Opie::Core::OConfig *cfg = Opie::Core::OApplication::oApplication()->config();
    Opie::Core::OConfig cfg("oggplay");
    Opie::Core::OConfigGroupSaver csaver(&cfg, "OggPlayer");

    setToolBarsMovable( false );

    QToolBar *menuBarHolder = new QToolBar( this );
    menuBarHolder->setHorizontalStretchable( true );
    
    QMenuBar *mb = new QMenuBar( menuBarHolder );
    QToolBar *tb = new QToolBar( this );

    QPopupMenu *fileMenu = new QPopupMenu( this );
    QAction *a;
    a = new QAction( tr("Clear playlist"), Opie::Core::OResource::loadPixmap("new"),
                  tr("New"), Qt::CTRL+Qt::Key_N, this, "clear_action" );
    a->addTo( fileMenu );
    connect(a, SIGNAL(activated()), playlist, SLOT(clear()));
    
    a = new QAction( tr("Add file or playlist"), Opie::Core::OResource::loadPixmap("fileopen"),
                  tr("Add"), Qt::CTRL+Qt::Key_O, this, "open_action" );
    connect(a, SIGNAL(activated()), this, SLOT(addToPlaylist()));
    a->addTo( fileMenu );
    a->addTo( tb);

    a = new QAction( tr("Add folder to playlist"), Opie::Core::OResource::loadPixmap("folder_open"),
                  tr("Add folder"), Qt::CTRL+Qt::Key_F, this, "dir_open_action" );
    connect(a, SIGNAL(activated()), this, SLOT(addDirToPlaylist()));
    a->addTo( fileMenu );

    fileMenu->insertSeparator();
    a = new QAction( tr("Save current playlist"), Opie::Core::OResource::loadPixmap("save"),
                  tr("Save As"), Qt::CTRL+Qt::Key_S, this, "save_action" );
    connect(a, SIGNAL(activated()), this, SLOT(savePlaylist()));
    a->addTo( fileMenu );
    fileMenu->insertSeparator();
    tb->addSeparator();
    a = new QAction( tr("Exit this application"), Opie::Core::OResource::loadPixmap("quit_icon"),
                  tr("Quit"), Qt::CTRL+Qt::Key_Q, this, "quit_action" );
    connect(a, SIGNAL(activated() ),
            qApp, SLOT(quit() ) );
    a->addTo( fileMenu );

    mb->insertItem(tr("File"), fileMenu );

    QPopupMenu *plMenu = new QPopupMenu( this );

    a = new QAction( tr("Previous title"), Opie::Core::OResource::loadPixmap("back"),
                  tr("Back"), Qt::Key_Left, this, "back_action" );
    a->addTo( plMenu );
    a->addTo( tb);
    connect(a, SIGNAL(activated()), this, SLOT(backward()));

    a = new QAction( tr("Next title"), Opie::Core::OResource::loadPixmap("forward"),
                  tr("Forward"), Qt::Key_Right, this, "forward_action" );
    a->addTo( plMenu );
    a->addTo( tb);
    connect(a, SIGNAL(activated()), this, SLOT(forward()));

    plMenu->insertSeparator();
    a = new QAction( tr("Play selected title"), Opie::Core::OResource::loadPixmap("play2"),
                  tr("Play"), Qt::Key_Enter, this, "play_action" );
    a->addTo( plMenu );
    a->addTo( tb);
    connect(a, SIGNAL(activated()), player, SLOT(playpause()));
    QAccel *aa = new QAccel(this);
    aa->connectItem(aa->insertItem(Key_Space), player, SLOT(playpause()));
    
    a = new QAction( tr("Stop playing"), Opie::Core::OResource::loadPixmap("stop"),
                  tr("Stop"), Qt::Key_S, this, "stop_action" );
    a->addTo( plMenu );
    a->addTo( tb);
    connect(a, SIGNAL(activated()), player, SLOT(stop()));

    plMenu->insertSeparator();
    a = new QAction( tr("Repeat playlist"), Opie::Core::OResource::loadPixmap("oggplay/loop"),
                  tr("Loop"), Qt::Key_L, this, "loop_action" );
    a->setToggleAction(true);
    a->addTo( plMenu );
    a->addTo( tb);
    connect(a, SIGNAL(toggled(bool)), selector, SLOT(repeatToggled(bool)));
    connect(a, SIGNAL(toggled(bool)), this, SLOT(repeatToggled(bool)));
    a->setOn(cfg.readNumEntry("repeat",0));

    a = new QAction( tr("Play playlist randomly"), Opie::Core::OResource::loadPixmap("oggplay/shuffle"),
                  tr("Shuffle"), Qt::Key_R, this, "shuffle_action" );
    a->setToggleAction(true);
    a->addTo( plMenu );
    a->addTo( tb);
    connect(a, SIGNAL(toggled(bool)), selector, SLOT(shuffleToggled(bool)));
    connect(a, SIGNAL(toggled(bool)), this, SLOT(shuffleToggled(bool)));
    a->setOn(cfg.readNumEntry("shuffle",0));

    plMenu->insertSeparator();

    a = new QAction( tr("Increase volume"), Opie::Core::OResource::loadPixmap("up"),
                  tr("Louder"), Qt::Key_Up, this, "volume_up" );
    a->addTo( plMenu );
    connect(a, SIGNAL(activated()), this, SLOT(incVolume()));

    a = new QAction( tr("Decrease volume"), Opie::Core::OResource::loadPixmap("down"),
                  tr("Quieter"), Qt::Key_Down, this, "volume_down" );
    a->addTo( plMenu );
    connect(a, SIGNAL(activated()), this, SLOT(decVolume()));
    
    gainMenu = new QPopupMenu(this);
    gainMenu->setCheckable(true);
    gainMenu->insertItem(tr("None"), player, SLOT(selectNoGain()), 0, 0);
    gainMenu->insertItem(tr("Track"), player, SLOT(selectTrackGain()), 0, 1);
    gainMenu->insertItem(tr("Album"), player, SLOT(selectAlbumGain()), 0, 2);
    gaincfg = cfg.readNumEntry("replaygain",0);
    switch(gaincfg)
    {
	case 1:
	    player->selectTrackGain();
	    break;
	    
	case 2:
	    player->selectAlbumGain();
	    break;
	    
	default:
	    gaincfg = 0;
	    player->selectNoGain();
	    break;
    }
    gainMenu->setItemChecked(gaincfg, true);
    connect(gainMenu, SIGNAL(activated(int)),
	    this, SLOT(setGain(int)));

    plMenu->insertItem(Opie::Core::OResource::loadPixmap("volume"),
		    tr("ReplayGain"), gainMenu);
    
    plMenu->insertSeparator();

    screenSaverMenu = new QPopupMenu(this);
    screenSaverMenu->setCheckable(true);
    screenSaverMenu->insertItem(tr("never"), 0);
    screenSaverMenu->insertItem(tr("30 sec"), 300);
    screenSaverMenu->insertItem(tr("1 min"), 600);
    screenSaverMenu->insertItem(tr("2 min"), 1200);
    screenSaverMenu->insertItem(tr("5 min"), 3000);
    screenSaverMenu->insertItem(tr("10 min"), 6000);
    screenSaverMenu->insertItem(tr("30 min"), 18000);
    screenSaverMenu->insertItem(tr("60 min"), 36000);
    screenSaver = cfg.readNumEntry("screensaver",0);
    screenSaverMenu->setItemChecked(screenSaver, true);

    connect(screenSaverMenu, SIGNAL(activated(int)),
	    this, SLOT(setSaver(int)));
    
    plMenu->insertItem(Opie::Core::OResource::loadPixmap("wait"),
		    tr("Screensaver"), screenSaverMenu);

    a = new QAction( tr("Lock the keyboard"), Opie::Core::OResource::loadPixmap("suspend"),
                  tr("Keyboard lock"), Qt::Key_K, this, "keyboard_lock" );
    a->setToggleAction(true);
    a->addTo( plMenu );
    connect(a, SIGNAL(toggled(bool)), this, SLOT(keyboardLock(bool)));
    connect(this, SIGNAL(keyboardGrabToggle(bool)), a, SLOT(setOn(bool)));

    plMenu->insertSeparator();
    
    mb->insertItem(tr("Player"), plMenu );
}

void MainWindow::shuffleToggled(bool val)
{
    Opie::Core::OConfig cfg("oggplay");
    Opie::Core::OConfigGroupSaver csaver(&cfg, "OggPlayer");
    cfg.writeEntry("shuffle", val);
}

void MainWindow::repeatToggled(bool val)
{
    Opie::Core::OConfig cfg("oggplay");
    Opie::Core::OConfigGroupSaver csaver(&cfg, "OggPlayer");
    cfg.writeEntry("repeat", val);
}

void MainWindow::changeVol(int part)
{
    if(hasAlsa)
	changeVolAlsa(part);
    if(!hasAlsa)
	changeVolOSS(part);
}

void MainWindow::changeVolAlsa(int part)
{
    snd_mixer_t* mixer;
    snd_mixer_elem_t *elem;
    snd_mixer_selem_id_t *selem_id;

    if(snd_mixer_open(&mixer, 0)<0)
    {
	hasAlsa = false;
	return;
    }

    if(snd_mixer_attach(mixer, "hw:0")<0 ||
       snd_mixer_selem_register(mixer, NULL, NULL)<0 ||
       snd_mixer_load(mixer)<0)
    {
	snd_mixer_close(mixer);
	hasAlsa = false;
	return;
    }

    snd_mixer_selem_id_alloca(&selem_id);
    snd_mixer_selem_id_set_name(selem_id, "PCM");
    elem = snd_mixer_find_selem(mixer, selem_id);

    if(elem && snd_mixer_selem_has_playback_volume(elem))
    {
        long int voll, volr, min, max;
	
	if(snd_mixer_selem_get_playback_volume_range(elem, &min, &max)<0 ||
           snd_mixer_selem_get_playback_volume(elem,SND_MIXER_SCHN_FRONT_RIGHT,&volr)<0 ||
           snd_mixer_selem_get_playback_volume(elem,SND_MIXER_SCHN_FRONT_LEFT,&voll)<0)
	{
	    snd_mixer_close(mixer);
	    return;
	}
	
	volr += (max-min)/part;
	voll += (max-min)/part;
	
	if(volr<min) volr = min;
	else if(volr>max) volr = max;
	if(voll<min) voll = min;
	else if(voll>max) voll = max;

	snd_mixer_selem_set_playback_volume(elem,SND_MIXER_SCHN_FRONT_RIGHT,volr);
	snd_mixer_selem_set_playback_volume(elem,SND_MIXER_SCHN_FRONT_LEFT,voll);
    }

    snd_mixer_close(mixer);
}

void MainWindow::changeVolOSS(int part)
{
    int mix, mask;
    struct
    {
	signed char l;
	signed char r;
    } vol;
    
    if((mix = ::open("/dev/mixer", O_RDWR))<0)
	return;

    if(ioctl(mix, SOUND_MIXER_READ_DEVMASK, &mask)>=0 &&
       (mask & SOUND_MASK_VOLUME) &&
       ioctl(mix, SOUND_MIXER_READ_VOLUME, &vol)>=0)
    {
	vol.l += 100/part;
	vol.r += 100/part;
	if(vol.l<0) vol.l = 0; else if(vol.l>100) vol.l = 100;
	if(vol.r<0) vol.r = 0; else if(vol.r>100) vol.r = 100;

	ioctl(mix, SOUND_MIXER_WRITE_VOLUME, &vol);
    }
    
    ::close(mix);
}

void MainWindow::incVolume()
{
    changeVol(20);
}

void MainWindow::decVolume()
{
    changeVol(-20);
}

bool MainWindow::hasAlsa = true;

#define FBIOBLANK             0x4611

/* VESA Blanking Levels */
#define VESA_NO_BLANKING      0
#define VESA_VSYNC_SUSPEND    1
#define VESA_HSYNC_SUSPEND    2
#define VESA_POWERDOWN        3

void MainWindow::setDisplay(bool on)
{
    int fd;
    static int state = true;

    timeCount = 0;

/*
    if(Opie::Core::ODevice::inst()->model() != Opie::Core::Model_Zaurus_SL5500)
	return;
*/
    
    if(state == on)
	return;

    state = on;
    
    if((fd = ::open("/dev/fb0",  O_RDWR)) >= 0 ||
       (fd = ::open("/dev/fb/0", O_RDWR)) >= 0)
    {
	::ioctl(fd, FBIOBLANK, on ? VESA_NO_BLANKING : VESA_POWERDOWN);
        ::close(fd);

	if(!on) // Turn of lights.
	    Opie::Core::ODevice::inst()->setDisplayStatus(on);
    }

    if(on)
        releaseKeyboard();
    else
        grabKeyboard();

    return;
}

void MainWindow::mousePressEvent(QMouseEvent* ev)
{
    setDisplay(true);
    QMainWindow::mousePressEvent(ev);
}

void MainWindow::keyPressEvent(QKeyEvent* ev)
{
    setDisplay(true);
    QMainWindow::keyPressEvent(ev);
}

void MainWindow::timerEvent(QTimerEvent *ev)
{
    if(screenSaver && isActiveWindow())
    {
	timeCount++;
	if(timeCount == screenSaver)
	    setDisplay(false);
    }
    else
	timeCount=0;

    QMainWindow::timerEvent(ev);
}

void MainWindow::keyboardLock(bool val)
{
    if(val)
    {
	QPEApplication::grabKeyboard();
	setWFlags(WStyle_Customize);
    }
    else
    {
	QPEApplication::ungrabKeyboard();
	clearWFlags(WStyle_Customize);
    }

    keybgrabbed = val;
}

void MainWindow::focusOutEvent(QFocusEvent* ev)
{
    if(keybgrabbed)
	keyboardLock(false);

    emit keyboardGrabToggle(false);
    setDisplay(true);
    
    QMainWindow::focusOutEvent(ev);
}

void MainWindow::setSaver(int val)
{
    timeCount = 0;

    screenSaverMenu->setItemChecked(screenSaver, false);
    screenSaver = val;
    screenSaverMenu->setItemChecked(screenSaver, true);

    Opie::Core::OConfig cfg("oggplay");
    Opie::Core::OConfigGroupSaver csaver(&cfg, "OggPlayer");
    cfg.writeEntry("screensaver", screenSaver);
}

void MainWindow::setGain(int val)
{
    for(int i=0; i<3; i++)
	gainMenu->setItemChecked(i, i==val);

    Opie::Core::OConfig cfg("oggplay");
    Opie::Core::OConfigGroupSaver csaver(&cfg, "OggPlayer");
    cfg.writeEntry("replaygain", val);
}
