/*****************************************************************************
 * Description: Selects the next file to be played                           *
 * Copyright:   2007 by Alexander Motzkau                                    *
 * License:     GPL V2                                                       *
 *****************************************************************************/

#include <set>

#include <stdlib.h>
#include <time.h>

#include "selector.h"
#include "playlist.h"

Selector::Selector() : QObject(), shuffle(false), repeat(false),
    played(), toplay(), lastplayed(NULL), lastplayedpos(-1),
    lastplaylistpos(-1), playlistsize(0)
{
    srand(time(NULL));
}

void Selector::entryAdded(int, PlaylistEntry* entry)
{
    toplay.push_back(entry);
    playlistsize++;
}

template<class container>
bool Selector::removeEntry(container &cont, PlaylistEntry* entry)
{
    for(typename container::iterator it = cont.begin(); it != cont.end(); it++)
    {
	if((*it) == entry)
	{
	    cont.erase(it);
	    return true;
	}
    }

    return false;
}

void Selector::unplayEntries(std::deque<PlaylistEntry*>::iterator start,
        std::deque<PlaylistEntry*>::iterator end)
{

    std::set<PlaylistEntry*> tempset;

    if(lastplayedpos >= start-played.begin() &&
	lastplayedpos < end-played.begin())
	    lastplayedpos=-1;

    tempset.insert(start, end);
    played.erase(start, end);
    
    for(std::deque<PlaylistEntry*>::iterator it = played.begin();
	it != played.end(); it++)
    {
	std::set<PlaylistEntry*>::iterator setit = tempset.find(*it);
	if(setit != tempset.end())
	    tempset.erase(setit);
    }
    
    toplay.insert(toplay.end(), tempset.begin(), tempset.end());
}

void Selector::entryRemoved(int pos, PlaylistEntry* entry)
{
    if(pos<lastplaylistpos)
	lastplaylistpos--;
    else if(pos==lastplaylistpos)
	lastplaylistpos=-1;

    playlistsize--;
    
    if(!removeEntry(toplay, entry) && removeEntry(played, entry) &&
	entry == lastplayed)
    {
	lastplayed = NULL;
	lastplayedpos = -1;
    }
}

void Selector::entryMoved(int oldpos, PlaylistEntry*, int newpos)
{
    if(lastplaylistpos == oldpos)
	lastplaylistpos = newpos;
    else if(oldpos < newpos)
    {
	if(oldpos < lastplaylistpos && lastplaylistpos <= newpos)
	    lastplaylistpos--;
    }
    else if(oldpos > newpos)
    {
	if(oldpos > lastplaylistpos && lastplaylistpos >= newpos)
	    lastplaylistpos++;
    }
}

void Selector::songSelected(int pos, PlaylistEntry* entry)
{
    if(shuffle)
    {
	if(entry != lastplayed)
	{
	    if(lastplayedpos>=0)
		unplayEntries(
		    played.begin()+lastplayedpos,
		    played.end());
	    
	    removeEntry(toplay, entry);
	    played.push_back(entry);
	    lastplayedpos = played.size()-1;
	}
    }

    lastplayed = entry;
    lastplaylistpos = pos;
}

void Selector::forward(bool waitforend)
{
    if(shuffle)
    {
	if(lastplayedpos<0 && lastplayed)
	{
	    played.push_back(lastplayed);
	    removeEntry(toplay, lastplayed);
	}
	    
	if(repeat && played.size() >= toplay.size())
	    unplayEntries(played.begin(), played.begin()+(played.size()-toplay.size()+1)/2);
	
	if(lastplayedpos>=0 && lastplayedpos < played.size()-1)
	{
	    lastplayedpos++;
	    lastplayed = played[lastplayedpos];
	    emit choose(lastplayed, waitforend);
	}
	else if(toplay.size()>0)
	{
	    unsigned int pos = (((unsigned long long)toplay.size())*rand())/(RAND_MAX+1ULL);
	    lastplayed = toplay[pos];
	    toplay.erase(toplay.begin()+pos);
	    lastplayedpos = played.size();
	    played.push_back(lastplayed);
	    
	    emit choose(lastplayed, waitforend);
	}
    }
    else
    {
	if(played.size()>0)
	    unplayEntries(played.begin(), played.end());
	
	if(lastplaylistpos<playlistsize-1)
	    emit choose(lastplaylistpos+1, waitforend);
	else if(repeat && playlistsize>0)
	    emit choose(0, waitforend);
    }
}

void Selector::backward(bool waitforend)
{
    if(shuffle)
    {
	if(lastplayedpos<0 && lastplayed)
	{
	    played.push_front(lastplayed);
	    removeEntry(toplay, lastplayed);
	}

	if(repeat && played.size() >= toplay.size())
	    unplayEntries(played.end()-(played.size()-toplay.size()+1)/2, played.end());
	
	if(lastplayedpos > 0)
	{
	    lastplayedpos--;
	    lastplayed = played[lastplayedpos];
	    emit choose(lastplayed, waitforend);
	}
	else if(toplay.size()>0)
	{
	    unsigned int pos = (((unsigned long long)toplay.size())*rand())/(RAND_MAX+1ULL);
	    lastplayed = toplay[pos];
	    toplay.erase(toplay.begin()+pos);
	    lastplayedpos = 0;
	    played.push_front(lastplayed);
	    
	    emit choose(lastplayed, waitforend);
	}
    }
    else
    {
	if(played.size()>0)
	    unplayEntries(played.begin(), played.end());
	
	if(lastplaylistpos>0)
	    emit choose(lastplaylistpos-1, waitforend);
	else if(repeat && playlistsize>0)
	    emit choose(playlistsize-1, waitforend);
    }
}

void Selector::shuffleToggled(bool val)
{
    shuffle = val;
}

void Selector::repeatToggled(bool val)
{
    repeat = val;
}
